(* Syntax.t -> FlatSyntax.t *)

open Syntax
open FlatSyntax

let find x env =
  try M2.find x env with Not_found -> failwith ("unbound variable " ^ x)

(* let のネストを多少減らす *)
let rec insert xs e1 e2 =
  match e1 with
  | Let(ys, e3, e4) -> Let(ys, e3, insert xs e4 e2)
  | LetRec(y, zs, e3, e4) -> LetRec(y, zs, e3, insert xs e4 e2)
  | _ -> Let(xs, e1, e2)

(* let を挿入する補助関数 *)
let insert_let e k =
  match e with
  | V v -> k v
  | _ -> let x = Id.new_id "var" in
    insert [x] e (k (Var x))

(* env は string から Id.t への写像 *)
let rec g env e =
  match e with
  | Syntax.Unit -> V Unit
  | Syntax.Int i -> V(Int i)
  | Syntax.Float f -> V(Float f)
  | Syntax.Bool b -> V(Bool b)
  | Syntax.Var x -> V(Var(find x env))
  | Syntax.Op(e1, e2s) ->
      insert_let (g env e1)
        (fun x -> let rec iter es ys_r =
           match es with
           | [] -> Op(x, List.rev ys_r)
           | e::es -> insert_let (g env e) (fun y -> iter es (y::ys_r))
         in
         iter e2s [])
  | Syntax.Tuple es ->
      let rec iter es xs_r =
        match es with
        | [] -> Tuple(List.rev xs_r)
        | e::es -> insert_let (g env e) (fun x -> iter es (x::xs_r))
      in
      iter es []
  | Syntax.If(e1, e2, e3) -> (* 後で最適化しないとひどいかも *)
      insert_let (g env e1)
        (fun x -> If(Eq, x, Bool false, g env e3, g env e2))
  | Syntax.Let((xs, e1), e2) ->
      let xs_ = List.map Id.new_id xs in
      let env2 = M2.add_list2 xs xs_ env in
      insert xs_ (g env e1) (g env2 e2)
  | Syntax.LetRec((f, xs, e1), e2) ->
      let f_ = Id.new_id f in
      let xs_ = List.map Id.new_id xs in
      let env2 = M2.add f f_ env in
      let env1 = M2.add_list2 xs xs_ env2 in
      LetRec(f_, xs_, g env1 e1, g env2 e2)
  | LetExt((x,t), e) ->
      match t with
      | Type.Tfun(ts,_) | Type.ForAll(_,Type.Tfun(ts,_)) ->
          let ys = List.map (Id.new_id_t "extarg") ts in
          let x1 = Id.new_id_t x t in
          let x2 = Id.special_id_t x t in
          let vs = List.map (fun y -> Var y) ys in
          let env_ = M2.add x x1 env in
          LetRec(x1, ys, ExtApp(x2, vs), g env_ e)
      | _ -> assert false

let f e =
  let env = List.fold_left
      (fun m x -> M2.add x.Id.name x m)
      M2.empty
      Builtin.fun_idents
  in
  g env e
